;; Generated package description from show-font.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "show-font" "1.0.0.0.20260111.105401" "Show font features in a buffer" '((emacs "29.1")) :commit "0fb3b3a7dbec93345c7135847b467c32cc45cbfa" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :keywords '("convenience" "writing" "font") :url "https://github.com/protesilaos/show-font")
